//#####################################################################
// Copyright 2002-2005, Robert Bridson, Eran Guendelman, Eilene Hao, Geoffrey Irving, Joseph Teran.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Viewer/VIEWER_OPTIONS.h>
#include "../QT_RENDER_UTILS.h"
#include "QT_AXES.h"
#include "QT_COMPONENT_OPTIONS.h"
#include <QAction>
namespace PhysBAM{
#define SET(name,value) (opt.name.Set(value),raw_opt[0].name=opt.name)
#define GET(name) opt.name.Value()
//#####################################################################
// Constructor
//#####################################################################
template<int d> QT_AXES<d>::
QT_AXES(QT_WORLD& world,const std::string& name)
    :QT_COMPONENT(world,name),opt(OPT::def),raw_opt{{},{},{},OPT::def}
{
}
//#####################################################################
// Constructor
//#####################################################################
template<int d> QT_AXES<d>::
~QT_AXES()
{
}
//#####################################################################
// Function Display
//#####################################################################
template<int d> void QT_AXES<d>::
Render(QT_RENDER_UTILS& ru)
{
    if(!GET(draw)) return;
    Read();

    if(ru.select) return;
    VECTOR<TV,2*TV::m> pts;
    VECTOR<COLOR,2*TV::m> cols;
    for(int a=0;a<TV::m;a++){
        COLOR col=COLOR()+(T).25;
        col(a)=1;
        pts(2*a)=X;
        pts(2*a+1)=X.Add_Axis(a,GET(length));
        cols(2*a)=col;
        cols(2*a+1)=col;}
    ru.Render_Lines<TV>(pts,cols);
}
//#####################################################################
// Function Bounding_Box
//#####################################################################
template<int d> auto QT_AXES<d>::
Bounding_Box() -> RANGE<FV>
{
    if(!GET(draw)) return RANGE<FV>::Empty_Box();
    Read();
    return {FV(X),FV(X+GET(length))};
}
//#####################################################################
// Function Read
//#####################################################################
template<int d> void QT_AXES<d>::
Read()
{
}
//#####################################################################
// Function Invalidate
//#####################################################################
template<int d> void QT_AXES<d>::
Invalidate(bool invalidate_common)
{
}
//#####################################################################
// Function Clear_Selection
//#####################################################################
template<int d> void QT_AXES<d>::
Clear_Selection()
{
}
//#####################################################################
// Function Set_Selection
//#####################################################################
template<int d> void QT_AXES<d>::
Set_Selection(ARRAY_VIEW<int> selection)
{
    PHYSBAM_WARNING("TODO: not implemented");
}
//#####################################################################
// Function Selection_Point
//#####################################################################
template<int d> auto QT_AXES<d>::
Selection_Point() -> FV
{
    PHYSBAM_WARNING("TODO: not implemented");
    return FV();
}
//#####################################################################
// Function Scale
//#####################################################################
template<int d> void QT_AXES<d>::
Scale(const T scale)
{
    SET(length,GET(length)*scale);
}
//#####################################################################
// Function Make_Dialog
//#####################################################################
template<int d> void QT_AXES<d>::
Make_Dialog(QT_WINDOW& window)
{
    QT_COMPONENT_OPTIONS* dialog=new QT_COMPONENT_OPTIONS(window,*this);
    dialog->setWindowTitle(QString::fromStdString(name));
    OPT opt2,raw_opt2;
    opt2=opt;
    raw_opt2=raw_opt[0];
    dialog->Add("Show object in viewer",opt.draw,raw_opt[0].draw);
    dialog->Add("Name to display in viewer",opt.name,raw_opt[0].name);
    dialog->Add("Axis length",opt.length,raw_opt[0].length);
    dialog->Add_End_Buttons();
    int result=dialog->exec();
    if(result!=QDialog::Accepted)
    {
        opt=opt2;
        raw_opt[0]=raw_opt2;
    }
    window.update();
}
//#####################################################################
// Function Read_Option
//#####################################################################
template<int d> void QT_AXES<d>::
Read_Option(std::istream& in,int level)
{
    auto& o=raw_opt[level];
    o.Read_Entry(in);
    opt=raw_opt[3];
    opt.Merge(raw_opt[2]);
    opt.Merge(raw_opt[1]);
    opt.Merge(raw_opt[0]);
}
//#####################################################################
// Function Set_Draw
//#####################################################################
template<int d> void QT_AXES<d>::
Set_Draw(bool draw)
{
    SET(draw,draw);
    action->setChecked(Draw());
}
//#####################################################################
// Function Draw
//#####################################################################
template<int d> bool QT_AXES<d>::
Draw() const
{
    return GET(draw);
}
//#####################################################################
// Function Save_Settings
//#####################################################################
template<int d> void QT_AXES<d>::
Save_Settings(std::ostream& out)
{
    raw_opt[0].Write(out,name);
}
//#####################################################################
// Function Clear_Settings
//#####################################################################
template<int d> void QT_AXES<d>::
Clear_Settings()
{
    raw_opt[0]={};
    opt=raw_opt[3];
    opt.Merge(raw_opt[2]);
    opt.Merge(raw_opt[1]);
    opt.Merge(raw_opt[0]);
}//#####################################################################
template class QT_AXES<2>;
template class QT_AXES<3>;
}
