//#####################################################################
// Copyright 2020, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Grids/GRID.h>
#include <Core/Math_Tools/RANGE.h>
#include "../QT_WORLD.h"
#include "QT_GRID.h"
#include "QT_GRID_COMPONENT.h"
namespace PhysBAM{
//#####################################################################
// Constructor
//#####################################################################
template<int d> QT_GRID_COMPONENT<d>::
QT_GRID_COMPONENT(QT_WORLD& world,const std::string& name,const std::string& grid_file)
    :QT_COMPONENT(world,name),
    ptr(dynamic_pointer_cast<QT_GRID<TV::m>>(world.Find_Component(grid_file))),
    grid(ptr->Grid())
{
    ptr->Add_Object(*this);
    this->selectable=false;
}
//#####################################################################
// Destructor
//#####################################################################
template<int d> QT_GRID_COMPONENT<d>::
~QT_GRID_COMPONENT()
{
    ptr->Remove_Object(*this);
}
//#####################################################################
// Function Set_Selection
//#####################################################################
template<int d> void QT_GRID_COMPONENT<d>::
Set_Selection(ARRAY_VIEW<int> selection)
{
    PHYSBAM_FATAL_ERROR("attempting to select a grid component");
}
//#####################################################################
// Function Clear_Selection
//#####################################################################
template<int d> void QT_GRID_COMPONENT<d>::
Clear_Selection()
{
}
//#####################################################################
// Function Selection_Point
//#####################################################################
template<int d> auto QT_GRID_COMPONENT<d>::
Selection_Point() -> FV
{
    PHYSBAM_FATAL_ERROR("attempting to get selection point on a grid component");
}
//#####################################################################
// Function Bounding_Box_Helper
//#####################################################################
template<int d> auto QT_GRID_COMPONENT<d>::
Bounding_Box_Helper(const RANGE<IV>& indices) const -> RANGE<FV>
{
    TV A=grid.Node_X(indices.min_corner);
    TV B=grid.Node_X(indices.max_corner);
    return RANGE<FV>(FV(A),FV(B));
}
//#####################################################################
// Function Grid_Name
//#####################################################################
template<int d> std::string QT_GRID_COMPONENT<d>::
Grid_Name() const
{
    return ptr->Name();
}
template class QT_GRID_COMPONENT<2>;
template class QT_GRID_COMPONENT<3>;
}
