//#####################################################################
// Copyright 2012.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Analytic_Tests/ANALYTIC_SCALAR.h>
#include <Core/Diff_Tools/Diff_Test.h>
#include <Core/Random_Numbers/RANDOM_NUMBERS.h>
#include <Core/Symbolics/PROGRAM_COMPILER.h>

namespace PhysBAM{

//#####################################################################
// Destructor
//#####################################################################
template<class TV> ANALYTIC_SCALAR<TV>::
~ANALYTIC_SCALAR()
{
}
//#####################################################################
// Function L
//#####################################################################
template<class TV> auto ANALYTIC_SCALAR<TV>::
L(const TV& X,T t) const -> T
{
    return ddX(X,t).Trace();
}
//#####################################################################
// Function Test
//#####################################################################
template<class TV> void ANALYTIC_SCALAR<TV>::
Test(const TV& X,T t) const
{
    TV dx;
    T eps=(T)1e-6,ds;
    RANDOM_NUMBERS<T> rand;
    rand.Set_Seed(1234);
    rand.Fill_Uniform(dx,-eps,eps);
    rand.Fill_Uniform(ds,-eps,eps);
    TV Y=X+dx;
    T s=t+ds;

    Diff_Test("analytic velocity x",f(X,t),X,dX(X,t),f(Y,t),Y,dX(Y,t));
    Diff_Test("analytic velocity dx",dX(X,t),X,ddX(X,t),dX(Y,t),Y,ddX(Y,t));
    Diff_Test("analytic velocity t",f(X,t),t,dt(X,t),f(X,s),s,dt(X,s));
}
//#####################################################################
// Constructor
//#####################################################################
template<class TV> ANALYTIC_SCALAR_PROGRAM<TV>::
ANALYTIC_SCALAR_PROGRAM(const std::string& str,const std::string& out_name)
    :h(str,out_name,TV::m,-1,-1,true)
{
}
//#####################################################################
// Destructor
//#####################################################################
template<class TV> ANALYTIC_SCALAR_PROGRAM<TV>::
~ANALYTIC_SCALAR_PROGRAM()
{
}
//#####################################################################
// Function f
//#####################################################################
template<class TV> auto ANALYTIC_SCALAR_PROGRAM<TV>::
f(const TV& X,T t) const -> T
{
    h.f(X,t);
    return h.context.data_out(h.idx_f(0));
}
//#####################################################################
// Function dX
//#####################################################################
template<class TV> TV ANALYTIC_SCALAR_PROGRAM<TV>::
dX(const TV& X,T t) const
{
    h.dX(X,t);
    return TV(h.context_dx.data_out.Subset(h.idx_dx));
}
//#####################################################################
// Function ddX
//#####################################################################
template<class TV> auto ANALYTIC_SCALAR_PROGRAM<TV>::
ddX(const TV& X,T t) const -> SYMMETRIC_MATRIX<T,TV::m> 
{
    h.ddX(X,t);
    SYMMETRIC_MATRIX<T,TV::m> out;
    for(int j=0;j<TV::m;j++)
        for(int k=j;k<TV::m;k++)
            out(j,k)=h.context_ddx.data_out(h.idx_ddx(j+TV::m*k));
    return out;
}
//#####################################################################
// Function dt
//#####################################################################
template<class TV> auto ANALYTIC_SCALAR_PROGRAM<TV>::
dt(const TV& X,T t) const -> T
{
    h.dt(X,t);
    return h.context_dt.data_out(h.idx_dt(0));
}
template struct ANALYTIC_SCALAR<VECTOR<double,2> >;
template struct ANALYTIC_SCALAR<VECTOR<double,3> >;
template struct ANALYTIC_SCALAR<VECTOR<float,2> >;
template struct ANALYTIC_SCALAR<VECTOR<float,3> >;
template struct ANALYTIC_SCALAR_PROGRAM<VECTOR<double,2> >;
template struct ANALYTIC_SCALAR_PROGRAM<VECTOR<double,3> >;
template struct ANALYTIC_SCALAR_PROGRAM<VECTOR<float,2> >;
template struct ANALYTIC_SCALAR_PROGRAM<VECTOR<float,3> >;
}
