//#####################################################################
// Copyright 2012.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Analytic_Tests/ANALYTIC_VECTOR.h>
#include <Core/Diff_Tools/Diff_Test.h>
#include <Core/Random_Numbers/RANDOM_NUMBERS.h>
#include <Core/Symbolics/PROGRAM_COMPILER.h>

namespace PhysBAM{

//#####################################################################
// Destructor
//#####################################################################
template<class TV,int m> ANALYTIC_VECTOR<TV,m>::
~ANALYTIC_VECTOR()
{
}
//#####################################################################
// Function L
//#####################################################################
template<class TV,int m> auto ANALYTIC_VECTOR<TV,m>::
L(const TV& X,T t) const -> VECTOR<T,m>
{
    return Contract<1,2>(ddX(X,t));
}
//#####################################################################
// Function Test
//#####################################################################
template<class TV,int m> void ANALYTIC_VECTOR<TV,m>::
Test(const TV& X,T t) const
{
    TV dx;
    T eps=(T)1e-6,ds;
    RANDOM_NUMBERS<T> rand;
    rand.Set_Seed(1234);
    rand.Fill_Uniform(dx,-eps,eps);
    rand.Fill_Uniform(ds,-eps,eps);
    TV Y=X+dx;
    T s=t+ds;

    Diff_Test("analytic velocity x",f(X,t),X,dX(X,t),f(Y,t),Y,dX(Y,t));
    Diff_Test("analytic velocity dx",dX(X,t),X,ddX(X,t),dX(Y,t),Y,ddX(Y,t));
    Diff_Test("analytic velocity t",f(X,t),t,dt(X,t),f(X,s),s,dt(X,s));
}
//#####################################################################
// Constructor
//#####################################################################
template<class TV,int m> ANALYTIC_VECTOR_PROGRAM<TV,m>::
ANALYTIC_VECTOR_PROGRAM(const std::string& str,const std::string& out_name)
    :h(str,out_name,TV::m,m,-1,true)
{
}
//#####################################################################
// Destructor
//#####################################################################
template<class TV,int m> ANALYTIC_VECTOR_PROGRAM<TV,m>::
~ANALYTIC_VECTOR_PROGRAM()
{
}
//#####################################################################
// Function v
//#####################################################################
template<class TV,int m> auto ANALYTIC_VECTOR_PROGRAM<TV,m>::
f(const TV& X,T t) const -> VECTOR<T,m>
{
    h.f(X,t);
    return TV(h.context.data_out.Subset(h.idx_f));
}
//#####################################################################
// Function dt
//#####################################################################
template<class TV,int m> TV ANALYTIC_VECTOR_PROGRAM<TV,m>::
dt(const TV& X,T t) const
{
    h.dt(X,t);
    return TV(h.context_dt.data_out.Subset(h.idx_dt));
}
//#####################################################################
// Function dX
//#####################################################################
template<class TV,int m> auto ANALYTIC_VECTOR_PROGRAM<TV,m>::
dX(const TV& X,T t) const -> MATRIX<T,m,TV::m>
{
    h.dX(X,t);
    MATRIX<T,m,TV::m> M;
    for(int i=0;i<m;i++)
        for(int j=0;j<TV::m;j++)
            M(i,j)=h.context_dx.data_out(h.idx_dx(m*j+i));
    return M;
}
//#####################################################################
// Function ddX
//#####################################################################
template<class TV,int m> auto ANALYTIC_VECTOR_PROGRAM<TV,m>::
ddX(const TV& X,T t) const -> SYMMETRIC_TENSOR<T,0,m,TV::m>
{
    h.ddX(X,t);
    SYMMETRIC_TENSOR<T,0,m,TV::m> S;
    for(int i=0;i<m;i++)
        for(int j=0;j<TV::m;j++)
            for(int k=j;k<TV::m;k++)
                S(i,j,k)=h.context_ddx.data_out(h.idx_ddx(TV::m*m*k+m*j+i));
    return S;
}
template struct ANALYTIC_VECTOR<VECTOR<double,2>,2>;
template struct ANALYTIC_VECTOR<VECTOR<double,3>,3>;
template struct ANALYTIC_VECTOR<VECTOR<float,2>,2>;
template struct ANALYTIC_VECTOR<VECTOR<float,3>,3>;
template struct ANALYTIC_VECTOR_PROGRAM<VECTOR<double,2>,2>;
template struct ANALYTIC_VECTOR_PROGRAM<VECTOR<double,3>,3>;
template struct ANALYTIC_VECTOR_PROGRAM<VECTOR<float,2>,2>;
template struct ANALYTIC_VECTOR_PROGRAM<VECTOR<float,3>,3>;
}
