#include <Core/Interpolation/CUBIC_MN_INTERPOLATION.h>
#include <Core/Arrays/ARRAY.h>
#include <Core/Vectors/VECTOR.h>
namespace PhysBAM{
//#####################################################################
// Constructor
//#####################################################################
template<class T> CUBIC_MN_INTERPOLATION<T>::
CUBIC_MN_INTERPOLATION()
{
    Set_Parameters();
}
//#####################################################################
// Destructor
//#####################################################################
template<class T> CUBIC_MN_INTERPOLATION<T>::
~CUBIC_MN_INTERPOLATION()
{
}
//#####################################################################
// Function Set_Parameters
//#####################################################################
template<class T> void CUBIC_MN_INTERPOLATION<T>::
Set_Parameters(const T b,const T c)
{
    m0=(12-9*b-6*c)/6;
    m1=(-18+12*b+6*c)/6;
    m3=(6-2*b)/6;
    n0=(-b-6*c)/6;
    n1=(6*b+30*c)/6;
    n2=(-12*b-48*c)/6;
    n3=(8*b+24*c)/6;
}
//#####################################################################
// Function Cubic_MN_Weights
//#####################################################################
template<class T> ARRAY<T> CUBIC_MN_INTERPOLATION<T>::
Cubic_MN_Weights(const T alpha) const
{
    T alpha2=alpha*alpha;
    T alpha3=alpha2*alpha;
    ARRAY<T> weights;
    weights.Append((T)(-alpha3/2.+alpha2-.5*alpha)); //f_k-1
    weights.Append((T)(3.*alpha3/2.-5*alpha2/2.+1.)); //f_k 
    weights.Append((T)(-3.*alpha3/2.+2*alpha2+alpha/2.)); //f_k+1
    weights.Append((T)(alpha3/2.-alpha2/2.));  //f_k+2
    for(int i=0;i<weights.m;i++) weights(i)=max((T)0.,min((T)1.,weights(i)));
    return weights;
}

template class CUBIC_MN_INTERPOLATION<float>;
template class CUBIC_MN_INTERPOLATION<double>;
}
